/*======================================================================

    serial_qt.c - A driver for Quatech's asychronous PCMCIA serial devices
    Copyright (C) 2001 Tim Gobeli - Quatech Inc.

    this program is free software; you cand redistribute it and/or modify 
    it under the terms of the GNU General Public License as  published by 
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.

    This program is distributed in the hope that it will be useful, but 
    WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
    General Publice License for more details.

    serial_qt.c 1.20 2004/08/25 9:40:30
        


    The initial developer of the original code is David A. Hinds
    <dahinds@users.sourceforge.net>.  Portions created by David A. Hinds
    are Copyright (C) 1999 David A. Hinds.  All Rights Reserved.

 
    
======================================================================*/

//#include <pcmcia/config.h>
//#include <pcmcia/k_compat.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/serial_core.h>
#include <linux/major.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include <asm/uaccess.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>
#include <pcmcia/cisreg.h>

#include <linux/serial_8250.h> //need this

#include "serial_qt.h"


#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
    #define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
#else
    #define DEBUG(n, args...)  printk(KERN_DEBUG args)

#endif

MODULE_AUTHOR("Tim Gobeli <tgobeli@quatech.com>");
MODULE_DESCRIPTION("PCMCIA Serial Port module for Quatech Adapters");
MODULE_LICENSE("GPL");

 #define CLK_X8  8
/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Bit map of interrupts to choose from */
static u_int irq_mask = 0xdeb8;
static int irq_list[4] = { -1};

/* Enable the speaker? */
static int do_sound = 1;

static int opr_flags = 0;

MODULE_PARM(irq_mask, "i");
MODULE_PARM(irq_list, "1-4i");
MODULE_PARM(do_sound, "i");
MODULE_PARM(opr_flags, "i");  //                        0 0 0 - no touch 
                              //                        0 0 1 - x 1
                              //                        0 1 0 - x 2
                              //                        0 1 1 - x 4
                              //                        1 0 0 - x 8
/*====================================================================*/


static dev_link_t *dev_list = NULL;



struct file_operations serialqt_pcmcia_fops = {
    ioctl:  ioctl_serial_pcmcia,

};

static int Adapter_Index = 0;
//void cs_error(client_handle_t handle, int func, int ret) ;

/*====================================================================*/
/* ioctl_serial_pcmcia function for control parent device configurations */


static int ioctl_serial_pcmcia (struct inode *innod, struct file *filp, unsigned int cmd,
                      unsigned long arg)
{

    unsigned    ud_QT_Baseaddress, err;
    unsigned    ucOPR_OrgValue, ucOPR_NewValue, uc_Value;
    int *p_Num_of_adapters, counts, index, *p_QMCR_Value, tmp_int;
    //struct pci_dev *dev = NULL;
    Identity_struct *p_Identity_of;
    int Identity_Index;
    dev_link_t *pTemp_dev_link;


    if (_IOC_TYPE(cmd) != SERIALQT_PCI_IOC_MAGIC)
        return -ENOTTY;
    if (_IOC_NR(cmd) > SERIALQT_IOC_MAXNR)
        return -ENOTTY;

    err = 0;
    switch (cmd)
    {
    case SERIALQT_WRITE_QOPR: 
        pTemp_dev_link = dev_list;
        index = arg >> 16;
        counts = 0;
        //Go through the whole list and find the  adapter
        while (pTemp_dev_link)
        {
            dbgprint(KERN_DEBUG"ioctl_serial_pci  arg = 0x%lx\n", arg);
            //Setting one port QOPR on multiport adapter sets them all  
            if (((serial_info_t *)(pTemp_dev_link->priv))->Adapter_Index == index)
            {
                //We've found the adapter we want, lets change the QOPR time setting accordingly
                ud_QT_Baseaddress = pTemp_dev_link->io.BasePort1;
                ucOPR_OrgValue =  SerialReadQOPR_Reg(ud_QT_Baseaddress);
                //Just mask off the last two bits of the QOPR register cause thats all 
                //that needs changed there for clock rate control
                ucOPR_NewValue = ucOPR_OrgValue & 0xfc;
                ucOPR_NewValue = ucOPR_NewValue | arg;
                SerialWriteQOPR_Reg(ud_QT_Baseaddress, FALSE, ucOPR_NewValue);
                //We're done
                break;
            } else
            {
                pTemp_dev_link = (dev_link_t *)pTemp_dev_link->next;
            }

        }  //end while
        break;



    case SERIALQT_WRITE_QMCR:
        err = -ENOTTY; //initialize as error so if we don't find this one we give
        pTemp_dev_link = dev_list;
        index = arg >> 16;

        //Go through the whole list and find the  adapter
        while (pTemp_dev_link)
        {
            dbgprint(KERN_DEBUG"ioctl_serial_pci  arg = 0x%lx\n", arg);
            if (((serial_info_t *)(pTemp_dev_link->priv))->Adapter_Index == index)
            {
                ud_QT_Baseaddress = pTemp_dev_link->io.BasePort1;
                //must set each ports QMCR individually
                for(index = 0; index < (((serial_info_t *)(pTemp_dev_link->priv))->multi); index++) 
                {
                    //We've found the adapter we want, lets change the QMCR time setting accordingly
                    ud_QT_Baseaddress = ud_QT_Baseaddress + 8*index;
                    ucOPR_NewValue = arg;

                    dbgprint(KERN_DEBUG"ioctl SERIALQT_WRITE_QMCR:  Base Addrss = 0x%x,  value = 0x%x\n", ud_QT_Baseaddress, ucOPR_NewValue);
                    SerialWriteQMCR_Reg(ud_QT_Baseaddress, ucOPR_NewValue );
                    err = 0;
                }
                //We're done
                break;
            } else
            {
                pTemp_dev_link = pTemp_dev_link->next;
            }

        }  //end while

        break;
    case SERIALQT_READ_QMCR:
        err = -ENOTTY; //initialize as error so if we don't find this one we give
        //an error.
        p_QMCR_Value =  (int *)arg;
        pTemp_dev_link = dev_list;
        index = arg >> 16;
        counts = 0;
        while (pTemp_dev_link)
        {

            if (((serial_info_t *)(pTemp_dev_link->priv))->Adapter_Index == index)
            {
                ud_QT_Baseaddress = pTemp_dev_link->io.BasePort1;
                uc_Value =  SerialReadQMCR_Reg(ud_QT_Baseaddress);
                err = put_user(uc_Value, p_QMCR_Value);
                break;   //We're done

            } else
            {
                pTemp_dev_link = pTemp_dev_link->next;
            }
        }


        break;

    case SERIALQT_GET_NUMOF_UNITS: 
        p_Num_of_adapters = (int *)arg;
        counts = 0;
        pTemp_dev_link = dev_list;
        while(pTemp_dev_link)
        {
            pTemp_dev_link = pTemp_dev_link->next;
            counts++;
        }

        dbgprint(KERN_DEBUG"ioctl_serial_pcmcia writing counts = %d", counts);

        err = put_user(counts, p_Num_of_adapters);


        break;
    case SERIALQT_GET_THIS_UNIT:
        counts = 0;
        dbgprint(KERN_DEBUG"In SERIALQT_GET_THIS_UNIT");

        p_Identity_of = (Identity_struct *)arg;
        get_user(Identity_Index, &p_Identity_of->index);

        dbgprint(KERN_DEBUG"SERIALQT_GET_THIS_UNIT Identity_Index= 0x%x\n", Identity_Index);

        err = -ENOTTY; //initialize as error so if we don't find this one we give
                       //an error.




        //-----------------------------------------------------------------------

        pTemp_dev_link = dev_list;
        while (pTemp_dev_link)
        {

            if (((serial_info_t *)(pTemp_dev_link->priv))->Adapter_Index == Identity_Index)
            {
                tmp_int = (int)(((serial_info_t *)(pTemp_dev_link->priv))->prodid);

                err = put_user(tmp_int, &p_Identity_of->n_identity);
                break;   //We're done

            } else
            {
                pTemp_dev_link = pTemp_dev_link->next;
            }
        }







        break;

    case SERIALQT_IS422_EXTENDED:
        err = -ENOTTY; //initialize as error so if we don't find this one we give
        dbgprint(KERN_DEBUG"SERIALQT_IS422_EXTENDED \n");
        //an error.
        pTemp_dev_link = dev_list;
        index = arg >> 16;

        while (pTemp_dev_link)
        {

            if (((serial_info_t *)(pTemp_dev_link->priv))->Adapter_Index == index)
            {
                dbgprint(KERN_DEBUG"Found Adapter \n");

                ud_QT_Baseaddress = pTemp_dev_link->io.BasePort1;
                if (SerialDoesQMCR_Exist(ud_QT_Baseaddress))
                {
                    err = 1;
                    break;   //no MCR so return -ENOTTY;
                }
                if (SerialDoes422_Exist(ud_QT_Baseaddress))
                {
                    err = 2;
                    break;
                }
                dbgprint(KERN_DEBUG"SERIALQT_IS422_EXTENDED err = 0\n");

                err = 0; //If we got this far then it is a 422/485 extended feature device
                break; 


            } else
            {
                pTemp_dev_link = pTemp_dev_link->next;
            }
        }
        break;




    default:
        err = -ENOTTY;




    } //End switch
    return err;
}


static void SerialWriteQOPR_Reg(unsigned Base_Address, bool internal, unsigned Arg){

    unsigned    ulQT_ORaddress, ulQT_LCRaddress;
    unsigned    ucLCR_OrgValue, ucOPR_OrgValue, ucOPR_NewValue;

    //We only allow the last two bits to change on the QOR register
    ulQT_ORaddress = Base_Address + 7;
    ulQT_LCRaddress = Base_Address + 3;
    //Set DLAB to access options register
    ucLCR_OrgValue = inb(ulQT_LCRaddress);
    dbgprint(KERN_DEBUG"Read LCR at = 0x%x value = 0x%x\n", ulQT_LCRaddress, ucLCR_OrgValue);


    outb(0xbf, ulQT_LCRaddress);
    dbgprint(KERN_DEBUG" Wrote LCR at = 0x%x value = 0x%x\n", ulQT_LCRaddress, 0xbf);

    ucOPR_OrgValue = inb(ulQT_ORaddress);
    //If this comes from user -Just mask off the last two bits of the QOPR register cause thats all 
    //that needs changed there for clock rate control
    if(internal == FALSE)
    {
         ucOPR_NewValue = ucOPR_OrgValue & 0xfc;
         ucOPR_NewValue = ucOPR_NewValue | Arg; 
         dbgprint(KERN_DEBUG"Read qopr at = 0x%x value = 0x%x\n", ulQT_ORaddress, ucOPR_OrgValue);
    }
    else
        ucOPR_NewValue = Arg;

    //Write new value to options register
    outb(ucOPR_NewValue, ulQT_ORaddress );

    //**** Lets read the value back for testing purposes **************//
    dbgprint(KERN_DEBUG"SerialWriteQOPR_Reg: Wrote qopr at = 0x%x value = 0x%x\n", ulQT_ORaddress, ucOPR_NewValue);

    // ucLCR_OrgValue = 0;
    //ucLCR_OrgValue = inb(ulQT_ORaddress);
    //dbgprint(KERN_DEBUG"Read qopr at = 0x%x value = 0x%x", ulQT_ORaddress, ucLCR_OrgValue);

    //*****************************************************************//

    //Restore origninal lCR value
    outb(ucLCR_OrgValue, ulQT_LCRaddress ); 



}

unsigned SerialReadQOPR_Reg(unsigned Base_Address){

    unsigned    ulQT_ORaddress, ulQT_LCRaddress;
    unsigned    ucLCR_OrgValue, ucOPR_Value;

    //We only allow the last two bits to change on the QOR register
    ulQT_ORaddress = Base_Address + 7;
    ulQT_LCRaddress = Base_Address + 3;
    //Set DLAB to access options register
    ucLCR_OrgValue = inb(ulQT_LCRaddress);
 

    outb(0xbf, ulQT_LCRaddress);
     ucOPR_Value = inb(ulQT_ORaddress);
    //Restore origninal lCR value
    outb(ucLCR_OrgValue, ulQT_LCRaddress ); 
    dbgprint(KERN_DEBUG" SerialReadQOPR_Reg:Read QOR at Base = 0x%x value = 0x%x\n", Base_Address, ucOPR_Value);

    return ucOPR_Value;  



}


static void SerialWriteQMCR_Reg(unsigned Base_Address, unsigned Arg){

    unsigned    ulQT_ORaddress, ulQT_LCRaddress, ulQT_MCRaddress;
    unsigned    ucLCR_OrgValue, ucOPR_OrgValue;

    //We only allow the last two bits to change on the QOR register
    ulQT_ORaddress = Base_Address + 7;
    ulQT_LCRaddress = Base_Address + 3;
    ulQT_MCRaddress = Base_Address + 4;
    //Set DLAB to access options register
    ucLCR_OrgValue = inb(ulQT_LCRaddress);

    outb(0xbf, ulQT_LCRaddress);

    ucOPR_OrgValue = inb(ulQT_ORaddress);
    //Set options register QLAB bit to access QMCR register
    outb(ucOPR_OrgValue | 0x10, ulQT_ORaddress);
    //Now write to the QMCR register
    outb(Arg, ulQT_MCRaddress);

    //Restore original contents of QOPR register
    outb(ucLCR_OrgValue, ulQT_LCRaddress );

    //Restore origninal lCR value
    outb(ucLCR_OrgValue, ulQT_LCRaddress ); 



}

static unsigned SerialReadQMCR_Reg(unsigned Base_Address){

    unsigned    ulQT_ORaddress, ulQT_LCRaddress, ulQT_MCRaddress;
    unsigned    ucLCR_OrgValue, ucOPR_OrgValue, ucMCR_Value;

    //We only allow the last two bits to change on the QOR register
    ulQT_ORaddress = Base_Address + 7;
    ulQT_LCRaddress = Base_Address + 3;
    ulQT_MCRaddress = Base_Address + 4;
    //Set DLAB to access options register
    ucLCR_OrgValue = inb(ulQT_LCRaddress);

    outb(0xbf, ulQT_LCRaddress);

    ucOPR_OrgValue = inb(ulQT_ORaddress);
    //Set options register QLAB bit to access QMCR register
    outb(ucOPR_OrgValue | 0x10, ulQT_ORaddress);
    //Now read the QMCR register
    ucMCR_Value = inb(ulQT_MCRaddress);

    //Restore original contents of QOPR register
    outb(ucOPR_OrgValue, ulQT_ORaddress );

    //Restore origninal lCR value
    outb(ucLCR_OrgValue, ulQT_LCRaddress ); 

    return ucMCR_Value;



}
// SerialDoesQMCR_Exist tests for existence of QMCR register and returns 0 valueif it does
// and nonzero if it doesn't 

static int SerialDoesQMCR_Exist(unsigned Base_Address)
{
    unsigned    ulQT_ORaddress, ulQT_LCRaddress, ulQT_MCRaddress;
    unsigned    ucLCR_OrgValue, uc_Value;
    int        Status = 0;

    //We only allow the last two bits to change on the QOR register
    ulQT_ORaddress = Base_Address + 7;
    ulQT_LCRaddress = Base_Address + 3;
    ulQT_MCRaddress = Base_Address + 4;
    //Set DLAB to access options register
    ucLCR_OrgValue = inb(ulQT_LCRaddress);
    //Writting a 0xbf to lcr will make bet 5 of QOR go high if QOR and QMCR exist
    outb(0xbf, ulQT_LCRaddress);
    uc_Value = inb(ulQT_ORaddress);
    if (!(uc_Value & 0x20))//Is bit 5 (IDX) set
    {
        dbgprint(KERN_DEBUG"QMCR bit b5 not set \n");
        Status = -ENOTTY;
    }

    if (Status == 0)
    {
        //So far so good now write something other than 0xbf to the LCR to set IDX to 0
        //and test
        outb(0x80, ulQT_LCRaddress);
        //IDX should be clear
        uc_Value = inb(ulQT_ORaddress);
        if (uc_Value & 0x20)
            Status = -ENOTTY;

    }
    outb(ucLCR_OrgValue, ulQT_LCRaddress); //Restore original value
    return Status;

}


static int SerialDoes422_Exist(unsigned Base_Address)
{
    int Status;
    unsigned ucMCR_OrgValue, uc_Value;
    ucMCR_OrgValue = SerialReadQMCR_Reg(Base_Address);
    //Essentially we're ging to write a value of 0xff to the QMCR register and 
    //if we read back a nonzero value, then its a 422 device otherwize its 422
    uc_Value = 0xff;
    SerialWriteQMCR_Reg(Base_Address, uc_Value);
    uc_Value = 0;
    uc_Value = SerialReadQMCR_Reg(Base_Address);
    if (uc_Value)
        Status = 0;
    else
        Status = -ENOTTY;

    SerialWriteQMCR_Reg(Base_Address, ucMCR_OrgValue);

    return Status;

}

/**************************************************************************
 *
 * TAG  3/24/2004  4:23pm
 *
 * FUNCTION:      SerialLockOutOverride
 *
 * DESCRIPTION:   We check on the presences of QOR and return true if we don't
 *                get the desired test pattern resuls
 *
 *
 * INPUTS:        Base address
 *
 *
 *
 * OUTPUTS:       None
 *
 *
 * RETURNS:       False if QOPR exist, TRUE if jumper overridden on Qopr isn't there *
 * NOTES:
 *
 *
 *************************************************************************/

static bool SerialLockOutOverride(unsigned Base_Address)
{
    unsigned QOPR_OrgValue, QOPR_Value, TestOutValue;


    QOPR_OrgValue =  SerialReadQOPR_Reg( Base_Address);

    QOPR_Value = QOPR_OrgValue & QPCR_TEST_FOR1; // Clear ID0, ID1 bits
    SerialWriteQOPR_Reg(Base_Address, TRUE, QOPR_Value);


    TestOutValue = SerialReadQOPR_Reg(Base_Address);
    TestOutValue &= 0xc0;  //Mask ID0 & ID1
    if (TestOutValue != QPCR_TEST_GET1)
        return TRUE;

    QOPR_Value = QOPR_OrgValue & QPCR_TEST_FOR1; // Clear ID0, ID1 bits
    QOPR_Value = QOPR_Value | QPCR_TEST_FOR2; // set ID0, ID1 bits

    SerialWriteQOPR_Reg(Base_Address, TRUE, QOPR_Value);
    TestOutValue = SerialReadQOPR_Reg(Base_Address);

    TestOutValue &= 0xc0;    //mask off ID0 and ID1 bits

    if (TestOutValue != QPCR_TEST_GET2)
        return TRUE;
    //----------------------------------------
    QOPR_Value = QOPR_OrgValue & QPCR_TEST_FOR1; // Clear ID0, ID1 bits
    QOPR_Value = QOPR_Value | QPCR_TEST_FOR3; // set ID0, ID1 bits

    SerialWriteQOPR_Reg(Base_Address, TRUE, QOPR_Value);
    TestOutValue = SerialReadQOPR_Reg(Base_Address);

    TestOutValue &= 0xc0;    //mask off ID0 and ID1 bits

    if (TestOutValue != QPCR_TEST_GET3)
        return TRUE;
    //----------------------------------------
    QOPR_Value = QOPR_OrgValue & QPCR_TEST_FOR1; // Clear ID0, ID1 bits
    QOPR_Value = QOPR_Value | QPCR_TEST_FOR4; // set ID0, ID1 bits

    SerialWriteQOPR_Reg(Base_Address, TRUE, QOPR_Value);
    TestOutValue = SerialReadQOPR_Reg(Base_Address);

    TestOutValue &= 0xc0;    //mask off ID0 and ID1 bits

    SerialWriteQOPR_Reg(Base_Address, TRUE, QOPR_OrgValue); //Restore value


    if (TestOutValue != QPCR_TEST_GET4)
        return TRUE;








    return FALSE;


}
/**************************************************************************
 *
 * TAG  4/28/2002  4:23pm
 *
 * FUNCTION:      SerialClkCtrlOverride
 *
 * DESCRIPTION:   Called from muli_config
 *                This function is called to determine if overrides to the Clock mode of the
 *                hardware feature have been set via jumper configurations
 *                also determines max clock multiplier and sets the adapter accordingly
 *                (some cards have a max multiplier of X4 some
 *                of X8
 *
 *
 *
 * INPUTS:        Base_Address        --- Adapter Base Address*
 *
 *
 * OUTPUTS:       None
 *
 *
 * RETURNS:       True if no QOPR or locked out, also appropriate clocke rate 
 *                is returned in the pClock_Rate parameter
 *
 * NOTES:
 *
 *
 *************************************************************************/

static bool SerialClkCtrlOverride(unsigned Base_Address, unsigned long *pClock_Rate)
{


    unsigned QOPR_OrgValue, QOPR_Value, TestOutValue, Auto_QOPR_Setting;
    bool status;

    status =  SerialLockOutOverride(Base_Address);
    //If no QOR we disable all these selections
    if (status == TRUE)
    {
        dbgprint(KERN_DEBUG"SerialClkCrtlOverride, no QOR\n"); 

        *pClock_Rate =  CLOCK_X1; //No Qor so use standard rate and return
        return status;
    }

    QOPR_OrgValue = SerialReadQOPR_Reg(Base_Address);
    QOPR_Value = QOPR_OrgValue & (~QOPR_CLOCK_X8); // Clear clk rate bits


    SerialWriteQOPR_Reg(Base_Address, FALSE, QOPR_Value);

    TestOutValue = SerialReadQOPR_Reg(Base_Address);
    //Those bits shoud be 0, if not we're overridden
    if (TestOutValue &= QOPR_CLOCK_X8)
    {
        *pClock_Rate =  CLOCK_X1; //Set to standard rate if overridden
        status =  TRUE;
    }
    else
    {
        //Now try setting them to "1" and see if any of it sticks
        // some cards only go X4 so only hi bit sticks
        QOPR_Value = QOPR_Value | QOPR_CLOCK_X8; // Set clk rate bits


        SerialWriteQOPR_Reg(Base_Address, TRUE, QOPR_Value);

        TestOutValue = SerialReadQOPR_Reg(Base_Address);
        if (TestOutValue &=QOPR_CLOCK_X8)
            status =  FALSE;  //It stuck, we're not overridden
        else
        {
            *pClock_Rate =  CLOCK_X1; //Set to standard rate if overridden
            status =  TRUE; //It didn't stick, we are overridden
        }

    }

    //Ok were not forced but the max rate is not the same for all boards
    //lets find out what the max is and  
    if (status == FALSE)
    {
        TestOutValue = SerialReadQOPR_Reg(Base_Address);
        TestOutValue &=QOPR_CLOCK_X8;
        switch (TestOutValue)
        {
        case QOPR_CLOCK_X2:
            *pClock_Rate =  CLOCK_X2;
            Auto_QOPR_Setting = QOPR_CLOCK_X2;
            break;

        case QOPR_CLOCK_X4:
            *pClock_Rate =  CLOCK_X4;
            Auto_QOPR_Setting = QOPR_CLOCK_X4;
            break;

        case QOPR_CLOCK_X8:
            *pClock_Rate =  CLOCK_X8;
            Auto_QOPR_Setting = QOPR_CLOCK_X8;
            break;
            break;
        default:
            *pClock_Rate =  CLOCK_X1;
             Auto_QOPR_Setting = QOPR_CLOCK_X1;
        }

        QOPR_OrgValue = QOPR_OrgValue & (~QOPR_CLOCK_MASK); //clear clock bits
        QOPR_OrgValue = QOPR_OrgValue | Auto_QOPR_Setting; //Set clock bits
        SerialWriteQOPR_Reg(Base_Address, TRUE,  QOPR_OrgValue);  //set adapter clock




    }
    SerialWriteQOPR_Reg(Base_Address, TRUE, QOPR_OrgValue);  //set adapter clock

    dbgprint(KERN_DEBUG"<1>SerialClkCtrlOverride:  Writing to QOPR value =  0x%04lx\n",(u_long)QOPR_OrgValue); 

    return status;


}



/*====================================================================*/
/*
 void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret};
    pcmcia_report_error(handle, &err);

}
*/

/*======================================================================

    serial_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/

static int serial_attach(struct pcmcia_device *p_dev)
{
    serial_info_t *info;
    dev_link_t *link;
    int i;

    DEBUG(0, "serial_attach()\n");

    /* Create new serial device */
    info = kmalloc(sizeof(*info), GFP_KERNEL);
    if (!info) return 0;
    memset(info, 0, sizeof(*info));
    link = &info->link; 
    link->priv = info;

    //link->release.function = &serial_release;

    link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
    link->io.NumPorts1 = 8;
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    if (irq_list[0] == -1)
        link->irq.IRQInfo2 = irq_mask;
    else
        for (i = 0; i < 4; i++)
            link->irq.IRQInfo2 |= 1 << irq_list[i];
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    if (do_sound)
    {
        link->conf.Attributes |= CONF_ENABLE_SPKR;
        link->conf.Status = CCSR_AUDIO_ENA;
    }
    link->conf.IntType = INT_MEMORY_AND_IO;

    link->handle = p_dev;

    p_dev->instance = link;
    link->next = dev_list;
    dev_list = link;

	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	serial_config(link);




    return 0;
} /* serial_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/




static void serial_detach(struct pcmcia_device *p_dev)
{
    dev_link_t *link = dev_to_instance(p_dev);
    serial_info_t *info = link->priv;
    dev_link_t **linkp, *linkq;
    int index;

    DEBUG(0, "serial_detach(0x%p)\n", link);


	/*
	 * Ensure any outstanding scheduled tasks are completed.
	 */
	flush_scheduled_work();

	/*
	 * Ensure that the ports have been released.
	 */
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
        if (*linkp == link) break;
    if (*linkp == NULL)
        return;

    if (link->state & DEV_CONFIG)
        serial_release(link);

         
    /* Unlink device structure, free bits */
    *linkp = link->next;
    kfree(info);

    //Now reindex each adpater
    index = 0;
    for(linkq = dev_list; linkq != NULL; linkq = linkq->next)
    {
        ((serial_info_t *)(linkq->priv))->Adapter_Index = index;
        index++;
    }
    //Update global adapter index
    Adapter_Index = index;
 

} /* serial_detach */

/*====================================================================*/

 /*====================================================================*/

static int setup_serial(client_handle_t handle, serial_info_t *info, kio_addr_t iobase, int irq, unsigned long clock_rate)
{
	struct uart_port port;
	int line, tmpe;

	memset(&port, 0, sizeof (struct uart_port));
	port.iobase = iobase;
	port.irq = irq;
	port.flags = UPF_BOOT_AUTOCONF | UPF_SKIP_TEST | UPF_SHARE_IRQ;
	port.uartclk = 1843200;
	port.dev = &handle_to_dev(handle);
	line = serial8250_register_port(&port);
	if (line < 0) {
		printk(KERN_NOTICE "serial_cs: serial8250_register_port() at "
		       "0x%04lx, irq %d failed\n", (u_long)iobase, irq);
		return -EINVAL;
	}

    info->line[info->ndev] = line;
	sprintf(info->node[info->ndev].dev_name, "ttyS%d", line);
	info->node[info->ndev].major = TTY_MAJOR;
	info->node[info->ndev].minor = 0x40 + line;
	if (info->ndev > 0)
		info->node[info->ndev - 1].next = &info->node[info->ndev];
	info->ndev++;

	return 0;
}

/*====================================================================*/

/*====================================================================*/
/*
static int get_tuple(int fn, client_handle_t handle, tuple_t *tuple,
                     cisparse_t *parse)
{
    int i;
    i = CardServices(fn, handle, tuple);
    if (i != CS_SUCCESS) return CS_NO_MORE_ITEMS;
    i = CardServices(GetTupleData, handle, tuple);
    if (i != CS_SUCCESS) return i;
    return CardServices(ParseTuple, handle, tuple, parse);
}   

*/
static int get_first_tuple(client_handle_t handle, tuple_t *tuple,
                     cisparse_t *parse)
{
    int i;
    i = pcmcia_get_first_tuple(handle, tuple);
    if (i != CS_SUCCESS) 
        return CS_NO_MORE_ITEMS;
     i = pcmcia_get_tuple_data(handle, tuple);
     if (i != CS_SUCCESS) 
         return i;
     return pcmcia_parse_tuple(handle, tuple, parse);
}

static int get_next_tuple(client_handle_t handle, tuple_t *tuple,
                     cisparse_t *parse)
{
    int i;
    i = pcmcia_get_next_tuple(handle, tuple);
    if (i != CS_SUCCESS) 
        return CS_NO_MORE_ITEMS;
     i = pcmcia_get_tuple_data(handle, tuple);
     if (i != CS_SUCCESS) 
         return i;
     return pcmcia_parse_tuple(handle, tuple, parse);
}


//#define first_tuple(a, b, c) get_tuple(GetFirstTuple, a, b, c)
#define first_tuple(a, b, c) get_first_tuple(a, b, c)

//#define next_tuple(a, b, c) get_tuple(GetNextTuple, a, b, c)
#define next_tuple(a, b, c) get_next_tuple(a, b, c)

/*====================================================================*/

static int multi_config(dev_link_t *link, int Adapter_Index)
{
    int status = 0;
    client_handle_t handle = link->handle;
    serial_info_t *info = link->priv;
    tuple_t tuple;
    u_char buf[256];
    cisparse_t parse;
    cistpl_cftable_entry_t *cf = &parse.cftable_entry;
    int i, base2 = 0;   
    unsigned long Clock_Rate;

    tuple.TupleData = (cisdata_t *)buf;
    tuple.TupleOffset = 0; tuple.TupleDataMax = 255;
    tuple.Attributes = 0;
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;

    /* First, look for a generic full-sized window */
    link->io.NumPorts1 = info->multi * 8;
    i = first_tuple(handle, &tuple, &parse);
    while (i != CS_NO_MORE_ITEMS)
    {
        if ((i == CS_SUCCESS) && (cf->io.nwin == 1))        
        {
            //TAG 09/23/03
            //Providing temorary way to get to the X8 mode
            if ((opr_flags == CLK_X8) && (info->extended))
            {
                printk("Setting index\n");
                cf->index |= 0x20;
            }
            link->conf.ConfigIndex = cf->index;
            link->io.BasePort1 = cf->io.win[0].base;
            link->io.IOAddrLines = cf->io.flags & CISTPL_IO_LINES_MASK;
            i = pcmcia_request_io( link->handle, &link->io);
            base2 = link->io.BasePort1 + 8;
            if (i == CS_SUCCESS) break;
        }
        i = next_tuple(handle, &tuple, &parse);
    }


    if (i != CS_SUCCESS)
    {
        cs_error(link->handle, RequestIO, i);
        return -1;
    }
     //link->irq.AssignedIRQ = 0x04;
    i = pcmcia_request_irq( link->handle, &link->irq);
    if (i != CS_SUCCESS)
    {
        cs_error(link->handle, RequestIRQ, i);
        link->irq.AssignedIRQ = 0;
    }

    i = pcmcia_request_configuration( link->handle, &link->conf);
    if (i != CS_SUCCESS)
    {
        cs_error(link->handle, RequestConfiguration, i);
        return -1;
    }

    /********************************************************************************/
    //SerialClkCtrlOverride will set the adapter clock rate to highest available rate and
    //return that rate in the varible

    SerialClkCtrlOverride((unsigned)link->io.BasePort1, &Clock_Rate); //W  

    /********************************************************************************/
    setup_serial(link->handle, info, link->io.BasePort1, link->irq.AssignedIRQ, Clock_Rate);
    /* The Nokia cards are not really multiport cards */

    for (i = 0; i < info->multi-1; i++)
        setup_serial(link->handle, info, base2+(8*i), link->irq.AssignedIRQ, Clock_Rate);

    //************************************************************
    //Filling in the lats node with device file name that accesses base device TAG 3/29/04
    status = 0; // Dynamic assignment of major number
    //store this for unregistering the device
    info->base_major_number = register_chrdev(status, "SerialQT_PCMCIA", &serialqt_pcmcia_fops);
    info->Adapter_Index = Adapter_Index;


    return 0;
}

/*======================================================================

    serial_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    serial device available to the system.

======================================================================*/

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn), args))!=0) goto cs_failed

static void serial_config(dev_link_t *link)
{
    client_handle_t handle = link->handle;
    serial_info_t *info = link->priv;
    tuple_t tuple;
    u_short buf[128];
    cisparse_t parse;
    int i, last_ret, last_fn;


    DEBUG(0, "serial_config(0x%p)\n", link);

    tuple.TupleData = (cisdata_t *)buf;
    tuple.TupleOffset = 0; tuple.TupleDataMax = 255;
    tuple.Attributes = 0;
    /* Get configuration register information */
    tuple.DesiredTuple = CISTPL_CONFIG;
    last_ret = first_tuple(handle, &tuple, &parse);
    if (last_ret != CS_SUCCESS)
    {
        last_fn = ParseTuple;
        goto cs_failed;
    }
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* Configure card */
    link->state |= DEV_CONFIG;

    /* Is this a compliant multifunction card? */
    tuple.DesiredTuple = CISTPL_LONGLINK_MFC;
    tuple.Attributes = TUPLE_RETURN_COMMON | TUPLE_RETURN_LINK;
    info->multi = (first_tuple(handle, &tuple, &parse) == CS_SUCCESS);

    /* Is this a multiport card? */
    tuple.DesiredTuple = CISTPL_MANFID;
    if (first_tuple(handle, &tuple, &parse) == CS_SUCCESS)
    {
        info->manfid = le16_to_cpu(buf[0]);
        for (i = 0; i < MULTI_COUNT; i++)
            if ((info->manfid == multi_id[i].manfid) &&
                (le16_to_cpu(buf[1]) == multi_id[i].prodid))
                break;
        if (i < MULTI_COUNT)
        {
            /*If we got here we got a match with the listed cards */
            info->multi = multi_id[i].multi;
            info->extended = multi_id[i].extended;
            info->prodid   = multi_id[i].prodid;
        }
    }


        multi_config(link, Adapter_Index);

    if (info->ndev == 0)
        goto failed;
    else
        Adapter_Index++;



    link->dev = &info->node[0];
    link->state &= ~DEV_CONFIG_PENDING;
    return;

    cs_failed:
    cs_error(link->handle, last_fn, last_ret);
    failed:
    serial_release(link);

} /* serial_config */

/*======================================================================

    After a card is removed, serial_release() will unregister the net
    device, and release the PCMCIA configuration.
    
======================================================================*/

static void serial_release(dev_link_t *link)
{
    //dev_link_t *link = (dev_link_t *)arg;
    serial_info_t *info = link->priv;
    int i;

    DEBUG(0, "serial_release(0x%p)\n", link);

    unregister_chrdev(info->base_major_number, "SerialQT_PCMCIA");
    for (i = 0; i < info->ndev; i++)
    {
/*        if( info->node[info->ndev].major == base_major_number)
        {
             unregister_chrdev(base_major_number, "SerialQT_PCMCIA");

        }
        else
*/
        //unregister_serial(info->line[i]);

        serial8250_unregister_port(info->line[i]);
    }
    link->dev = NULL;

    if (!info->slave)
    {
        pcmcia_release_configuration( link->handle);
        pcmcia_release_io(link->handle, &link->io);
        pcmcia_release_irq(link->handle, &link->irq);
    }

    link->state &= ~DEV_CONFIG;

} /* serial_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the serial drivers from
    talking to the ports.
    
======================================================================*/

static int serial_event(event_t event, int priority,
                        event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;
    serial_info_t *info = link->priv;

    DEBUG(1, "serial_event(0x%06x)\n", event);

    switch (event)
    {
    case CS_EVENT_CARD_REMOVAL:
        DEBUG(0, "serial_qt:CS_EVENT_CARD_REMOVAL \n");

        link->state &= ~DEV_PRESENT;
        if (link->state & DEV_CONFIG)
            serial_release(link);
        break;
    case CS_EVENT_CARD_INSERTION:
        link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
        serial_config(link);
        break;
    case CS_EVENT_PM_SUSPEND:
        link->state |= DEV_SUSPEND;
        /* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
        if ((link->state & DEV_CONFIG) && !info->slave)
            pcmcia_release_configuration(link->handle);
        break;
    case CS_EVENT_PM_RESUME:
        link->state &= ~DEV_SUSPEND;
        /* Fall through... */
    case CS_EVENT_CARD_RESET:
        if (DEV_OK(link) && !info->slave)
           pcmcia_request_configuration(link->handle, &link->conf);
        break;
    }
    return 0;
} /* serial_event */

/*====================================================================*/
//****************************
static struct pcmcia_driver serial_qt_driver = {
	.owner		= THIS_MODULE,
	.drv		= {
		.name	= "serial_qt",
	},
	.probe		= serial_attach,
	.remove		= serial_detach,
};

//**************************
static int __init init_serial_qt(void)
{
    return pcmcia_register_driver(&serial_qt_driver);
}

static void __exit exit_serial_qt(void)
{
    DEBUG(0, "serial_qt: unloading\n");
    pcmcia_unregister_driver(&serial_qt_driver);
}



module_init(init_serial_qt);
module_exit(exit_serial_qt);
